# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class UsageStatistics(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:

        class properties:

            class period(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'period':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class total_core_hours_failed(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'total_core_hours_failed':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class total_core_hours(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'total_core_hours':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class total_tasks(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.IntSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'total_tasks':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class total_failed_tasks(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.IntSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'total_failed_tasks':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class avg_total_tasks_duration(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'avg_total_tasks_duration':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class avg_computation_seconds_task_duration(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'avg_computation_seconds_task_duration':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class success_rate(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'success_rate':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class avg_estimated_cost(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'avg_estimated_cost':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class total_estimated_cost(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'total_estimated_cost':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class total_failed_cost(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'total_failed_cost':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            __annotations__ = {
                "period":
                    period,
                "total_core_hours_failed":
                    total_core_hours_failed,
                "total_core_hours":
                    total_core_hours,
                "total_tasks":
                    total_tasks,
                "total_failed_tasks":
                    total_failed_tasks,
                "avg_total_tasks_duration":
                    avg_total_tasks_duration,
                "avg_computation_seconds_task_duration":
                    avg_computation_seconds_task_duration,
                "success_rate":
                    success_rate,
                "avg_estimated_cost":
                    avg_estimated_cost,
                "total_estimated_cost":
                    total_estimated_cost,
                "total_failed_cost":
                    total_failed_cost,
            }

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["period"]
    ) -> MetaOapg.properties.period:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["total_core_hours_failed"]
    ) -> MetaOapg.properties.total_core_hours_failed:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["total_core_hours"]
    ) -> MetaOapg.properties.total_core_hours:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["total_tasks"]
    ) -> MetaOapg.properties.total_tasks:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["total_failed_tasks"]
    ) -> MetaOapg.properties.total_failed_tasks:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["avg_total_tasks_duration"]
    ) -> MetaOapg.properties.avg_total_tasks_duration:
        ...

    @typing.overload
    def __getitem__(
        self,
        name: typing_extensions.Literal["avg_computation_seconds_task_duration"]
    ) -> MetaOapg.properties.avg_computation_seconds_task_duration:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["success_rate"]
    ) -> MetaOapg.properties.success_rate:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["avg_estimated_cost"]
    ) -> MetaOapg.properties.avg_estimated_cost:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["total_estimated_cost"]
    ) -> MetaOapg.properties.total_estimated_cost:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["total_failed_cost"]
    ) -> MetaOapg.properties.total_failed_cost:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "period",
        "total_core_hours_failed",
        "total_core_hours",
        "total_tasks",
        "total_failed_tasks",
        "avg_total_tasks_duration",
        "avg_computation_seconds_task_duration",
        "success_rate",
        "avg_estimated_cost",
        "total_estimated_cost",
        "total_failed_cost",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["period"]
    ) -> typing.Union[MetaOapg.properties.period, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["total_core_hours_failed"]
    ) -> typing.Union[MetaOapg.properties.total_core_hours_failed,
                      schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["total_core_hours"]
    ) -> typing.Union[MetaOapg.properties.total_core_hours, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["total_tasks"]
    ) -> typing.Union[MetaOapg.properties.total_tasks, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["total_failed_tasks"]
    ) -> typing.Union[MetaOapg.properties.total_failed_tasks, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["avg_total_tasks_duration"]
    ) -> typing.Union[MetaOapg.properties.avg_total_tasks_duration,
                      schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self,
        name: typing_extensions.Literal["avg_computation_seconds_task_duration"]
    ) -> typing.Union[MetaOapg.properties.avg_computation_seconds_task_duration,
                      schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["success_rate"]
    ) -> typing.Union[MetaOapg.properties.success_rate, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["avg_estimated_cost"]
    ) -> typing.Union[MetaOapg.properties.avg_estimated_cost, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["total_estimated_cost"]
    ) -> typing.Union[MetaOapg.properties.total_estimated_cost, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["total_failed_cost"]
    ) -> typing.Union[MetaOapg.properties.total_failed_cost, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "period",
        "total_core_hours_failed",
        "total_core_hours",
        "total_tasks",
        "total_failed_tasks",
        "avg_total_tasks_duration",
        "avg_computation_seconds_task_duration",
        "success_rate",
        "avg_estimated_cost",
        "total_estimated_cost",
        "total_failed_cost",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        period: typing.Union[MetaOapg.properties.period, dict,
                             frozendict.frozendict, str, date, datetime,
                             uuid.UUID, int, float, decimal.Decimal, bool, None,
                             list, tuple, bytes, io.FileIO, io.BufferedReader,
                             schemas.Unset] = schemas.unset,
        total_core_hours_failed: typing.Union[
            MetaOapg.properties.total_core_hours_failed, dict,
            frozendict.frozendict, str, date, datetime, uuid.UUID, int, float,
            decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO,
            io.BufferedReader, schemas.Unset] = schemas.unset,
        total_core_hours: typing.Union[MetaOapg.properties.total_core_hours,
                                       dict, frozendict.frozendict, str, date,
                                       datetime, uuid.UUID, int, float,
                                       decimal.Decimal, bool, None, list, tuple,
                                       bytes, io.FileIO, io.BufferedReader,
                                       schemas.Unset] = schemas.unset,
        total_tasks: typing.Union[MetaOapg.properties.total_tasks, dict,
                                  frozendict.frozendict, str, date, datetime,
                                  uuid.UUID, int, float, decimal.Decimal, bool,
                                  None, list, tuple, bytes, io.FileIO,
                                  io.BufferedReader,
                                  schemas.Unset] = schemas.unset,
        total_failed_tasks: typing.Union[MetaOapg.properties.total_failed_tasks,
                                         dict, frozendict.frozendict, str, date,
                                         datetime, uuid.UUID, int, float,
                                         decimal.Decimal, bool, None, list,
                                         tuple, bytes, io.FileIO,
                                         io.BufferedReader,
                                         schemas.Unset] = schemas.unset,
        avg_total_tasks_duration: typing.Union[
            MetaOapg.properties.avg_total_tasks_duration, dict,
            frozendict.frozendict, str, date, datetime, uuid.UUID, int, float,
            decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO,
            io.BufferedReader, schemas.Unset] = schemas.unset,
        avg_computation_seconds_task_duration: typing.Union[
            MetaOapg.properties.avg_computation_seconds_task_duration, dict,
            frozendict.frozendict, str, date, datetime, uuid.UUID, int, float,
            decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO,
            io.BufferedReader, schemas.Unset] = schemas.unset,
        success_rate: typing.Union[MetaOapg.properties.success_rate, dict,
                                   frozendict.frozendict, str, date, datetime,
                                   uuid.UUID, int, float, decimal.Decimal, bool,
                                   None, list, tuple, bytes, io.FileIO,
                                   io.BufferedReader,
                                   schemas.Unset] = schemas.unset,
        avg_estimated_cost: typing.Union[MetaOapg.properties.avg_estimated_cost,
                                         dict, frozendict.frozendict, str, date,
                                         datetime, uuid.UUID, int, float,
                                         decimal.Decimal, bool, None, list,
                                         tuple, bytes, io.FileIO,
                                         io.BufferedReader,
                                         schemas.Unset] = schemas.unset,
        total_estimated_cost: typing.Union[
            MetaOapg.properties.total_estimated_cost, dict,
            frozendict.frozendict, str, date, datetime, uuid.UUID, int, float,
            decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO,
            io.BufferedReader, schemas.Unset] = schemas.unset,
        total_failed_cost: typing.Union[MetaOapg.properties.total_failed_cost,
                                        dict, frozendict.frozendict, str, date,
                                        datetime, uuid.UUID, int, float,
                                        decimal.Decimal, bool, None, list,
                                        tuple, bytes, io.FileIO,
                                        io.BufferedReader,
                                        schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'UsageStatistics':
        return super().__new__(
            cls,
            *_args,
            period=period,
            total_core_hours_failed=total_core_hours_failed,
            total_core_hours=total_core_hours,
            total_tasks=total_tasks,
            total_failed_tasks=total_failed_tasks,
            avg_total_tasks_duration=avg_total_tasks_duration,
            avg_computation_seconds_task_duration=
            avg_computation_seconds_task_duration,
            success_rate=success_rate,
            avg_estimated_cost=avg_estimated_cost,
            total_estimated_cost=total_estimated_cost,
            total_failed_cost=total_failed_cost,
            _configuration=_configuration,
            **kwargs,
        )



"""Class to run commands on Quantum Espresso."""
from typing import List, Optional
import logging

from inductiva import types, tasks, simulators


@simulators.simulator.mpi_enabled
class QuantumEspresso(simulators.Simulator):
    """Class to run commands on Quantum Espresso."""

    COMMANDS = [
        "alpha2f", "dvscf_q2r", "head", "matdyn", "plan_avg", "pw", "rism1d",
        "turbo_spectrum", "average", "dynmat", "hp", "molecularnexafs",
        "plotband", "pw2bgw", "scan_ibrav", "upfconv", "band_interpolation",
        "epa", "ibrav2cell", "molecularpdos", "plotproj", "pw2critic", "simple",
        "virtual_v2", "bands", "epsilon", "initial_state", "neb", "plotrho",
        "pw2gt", "simple_bse", "wannier90", "bse_main", "ev", "kcw",
        "open_grid", "pmw", "pw2gw", "simple_ip", "wannier_ham", "casino2upf",
        "fermi_proj", "kcwpp_interp", "oscdft_et", "postahc", "pw2wannier90",
        "spectra_correction", "wannier_plot", "cell2ibrav", "fermi_velocity",
        "kcwpp_sh", "oscdft_pp", "postw90", "pw4gww", "sumpdos", "wfck2r", "cp",
        "fqha", "kpoints", "path_interpolation", "pp", "pwcond",
        "turbo_davidson", "wfdd", "cppp", "fs", "lambda", "pawplot", "ppacf",
        "pwi2xsf", "turbo_eels", "xspectra", "d3hess", "gww", "ld1", "ph",
        "pprism", "q2qstar", "turbo_lanczos", "dos", "gww_fit", "manycp",
        "phcg", "projwfc", "q2r", "turbo_magnon"
    ]

    def __init__(self, /, version: Optional[str] = None, use_dev: bool = False):
        """Initialize the Quantum Espresso simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "quantumespresso"
        self.container_image = self._get_image_uri()

    @property
    def name(self):
        """Get the name of the this simulator."""
        return "Quantum-Espresso"

    def available_commands(self) -> List[str]:
        """Get the list of available commands for Quantum Espresso.
        Returns:
            List of available commands for Quantum Espresso.
        """
        logging.info("\nAll available commands for Quantum Espresso have two"
                     " versions.\nMPI version and OpenMP version.\n"
                     "For the OpenMP version, append `_openmp` to the command "
                     "names (e.g., pw_openmp.x).\n"
                     "For the MPI version, just use the normal command names "
                     " (e.g., pw.x).")
        return self.COMMANDS

    def run(self,
            input_dir: Optional[str],
            commands: List[str],
            *,
            storage_dir: Optional[str] = "",
            on: types.ComputationalResources,
            extra_metadata: Optional[dict] = None,
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[List[str]] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.
        Args:
            input_dir: Path to the directory containing the input files.
            commands: List of commands to run.
            on: The computatißonal resource to launch the simulation on.
            storage_dir: Parent directory for storing simulation
                               results.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
        """
        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           extra_metadata=extra_metadata,
                           container_image=self._image_uri,
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           **kwargs)



"""Register CLI command for login."""
import argparse
import getpass
import os

import inductiva
from inductiva import constants, users, utils


def login(args):
    """Prompts the user to enter their API Key and stores it securely."""

    # pylint: disable=trailing-whitespace,line-too-long
    inductiva_art = r"""     ___  _   _  ____   _   _   ____  _____  ___ __     __ _    
    |_ _|| \ | ||  _ \ | | | | / ___||_   _||_ _|\ \   / // \   
     | | |  \| || | | || | | || |      | |   | |  \ \ / // _ \  
     | | | |\  || |_| || |_| || |___   | |   | |   \ V // ___ \ 
    |___||_| \_||____/  \___/  \____|  |_|  |___|   \_//_/   \_\
    """
    print(inductiva_art)
    if os.path.exists(constants.API_KEY_FILE_PATH):
        print(
            "    You are already logged in. Run `inductiva auth logout` if you "
            "want to log out. \n"
            "    Setting a new API Key will erase the existing one.")

    prompt_func = getpass.getpass if args.private else input
    warning = " (input will not be visible)" if args.private else ""

    prompt = prompt_func(
        "    To log in, you need an API Key. You can obtain it "
        "from your account at https://console.inductiva.ai/account.\n"
        f"Please paste your API Key here{warning}: ")

    api_key = prompt.strip()

    # Set the API Key
    inductiva.set_api_key(api_key, login_message=False)

    # If the API Key is invalid, this will raise an exception
    user_info = users.get_info()

    utils.set_stored_api_key(api_key)

    user_name = user_info["name"] or ""
    print(f"Welcome back {user_name}!")


def register(parser):
    """Register the login command."""
    subparser = parser.add_parser("login",
                                  help="Login using Inductiva API Key.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "The `inductiva login` command allows you to log in using your key.\n"
        "You can obtain your API Key from your account at "
        "https://console.inductiva.ai/account.\n")

    subparser.add_argument("--private",
                           action="store_true",
                           help="Hide API Key.")

    subparser.set_defaults(func=login)



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class UserCostsDetails(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "costs",
            "expired",
            "tier_id",
            "email",
        }

        class properties:
            email = schemas.StrSchema
            tier_id = schemas.StrSchema
            expired = schemas.BoolSchema

            class costs(schemas.ListSchema):

                class MetaOapg:

                    @staticmethod
                    def items() -> typing.Type['CostDetail']:
                        return CostDetail

                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple['CostDetail'],
                                       typing.List['CostDetail']],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                ) -> 'costs':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )

                def __getitem__(self, i: int) -> 'CostDetail':
                    return super().__getitem__(i)

            class name(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'name':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            __annotations__ = {
                "email": email,
                "tier_id": tier_id,
                "expired": expired,
                "costs": costs,
                "name": name,
            }

    costs: MetaOapg.properties.costs
    expired: MetaOapg.properties.expired
    tier_id: MetaOapg.properties.tier_id
    email: MetaOapg.properties.email

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["email"]
    ) -> MetaOapg.properties.email:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["tier_id"]
    ) -> MetaOapg.properties.tier_id:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["expired"]
    ) -> MetaOapg.properties.expired:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["costs"]
    ) -> MetaOapg.properties.costs:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["name"]
    ) -> MetaOapg.properties.name:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "email",
        "tier_id",
        "expired",
        "costs",
        "name",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["email"]
    ) -> MetaOapg.properties.email:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["tier_id"]
    ) -> MetaOapg.properties.tier_id:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["expired"]
    ) -> MetaOapg.properties.expired:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["costs"]
    ) -> MetaOapg.properties.costs:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["name"]
    ) -> typing.Union[MetaOapg.properties.name, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "email",
        "tier_id",
        "expired",
        "costs",
        "name",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        costs: typing.Union[
            MetaOapg.properties.costs,
            list,
            tuple,
        ],
        expired: typing.Union[
            MetaOapg.properties.expired,
            bool,
        ],
        tier_id: typing.Union[
            MetaOapg.properties.tier_id,
            str,
        ],
        email: typing.Union[
            MetaOapg.properties.email,
            str,
        ],
        name: typing.Union[MetaOapg.properties.name, dict,
                           frozendict.frozendict, str, date, datetime,
                           uuid.UUID, int, float, decimal.Decimal, bool, None,
                           list, tuple, bytes, io.FileIO, io.BufferedReader,
                           schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'UserCostsDetails':
        return super().__new__(
            cls,
            *_args,
            costs=costs,
            expired=expired,
            tier_id=tier_id,
            email=email,
            name=name,
            _configuration=_configuration,
            **kwargs,
        )


from inductiva.client.model.cost_detail import CostDetail



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class FileDownloadDetails(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "file_path",
            "unzip",
            "url",
        }

        class properties:
            url = schemas.StrSchema
            file_path = schemas.StrSchema
            unzip = schemas.BoolSchema
            __annotations__ = {
                "url": url,
                "file_path": file_path,
                "unzip": unzip,
            }

    file_path: MetaOapg.properties.file_path
    unzip: MetaOapg.properties.unzip
    url: MetaOapg.properties.url

    @typing.overload
    def __getitem__(
            self,
            name: typing_extensions.Literal["url"]) -> MetaOapg.properties.url:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["file_path"]
    ) -> MetaOapg.properties.file_path:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["unzip"]
    ) -> MetaOapg.properties.unzip:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "url",
        "file_path",
        "unzip",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self,
            name: typing_extensions.Literal["url"]) -> MetaOapg.properties.url:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["file_path"]
    ) -> MetaOapg.properties.file_path:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["unzip"]
    ) -> MetaOapg.properties.unzip:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "url",
        "file_path",
        "unzip",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        file_path: typing.Union[
            MetaOapg.properties.file_path,
            str,
        ],
        unzip: typing.Union[
            MetaOapg.properties.unzip,
            bool,
        ],
        url: typing.Union[
            MetaOapg.properties.url,
            str,
        ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'FileDownloadDetails':
        return super().__new__(
            cls,
            *_args,
            file_path=file_path,
            unzip=unzip,
            url=url,
            _configuration=_configuration,
            **kwargs,
        )



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from inductiva.client.paths.compute_group.delete import DeleteVmGroup
from inductiva.client.paths.compute_group_status.get import GetGroupStatus
from inductiva.client.paths.compute_price.get import GetInstancePrice
from inductiva.client.paths.compute_group_name.get import GetVmGroupByName
from inductiva.client.paths.compute_groups.get import ListActiveUserInstanceGroups
from inductiva.client.paths.compute_machine_types.get import ListAvailableMachineTypes
from inductiva.client.paths.compute_groups_history.get import ListUserInstanceGroups
from inductiva.client.paths.compute_group.post import RegisterVmGroup
from inductiva.client.paths.compute_group_start.post import StartVmGroup


class ComputeApi(
        DeleteVmGroup,
        GetGroupStatus,
        GetInstancePrice,
        GetVmGroupByName,
        ListActiveUserInstanceGroups,
        ListAvailableMachineTypes,
        ListUserInstanceGroups,
        RegisterVmGroup,
        StartVmGroup,
):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    pass



"""Methods to interact with the user storage resources."""
import time
import logging
import os
import tqdm
from typing import List, Literal, Optional, Tuple
import urllib

import inductiva
from inductiva import constants
from inductiva.api import methods
from inductiva.client import exceptions, models
from inductiva.client.apis.tags import storage_api
from inductiva.utils import format_utils


def _print_storage_size_and_cost() -> int:
    """ Print the storage size and cost.

    return the storage size in bytes.
    """
    api = storage_api.StorageApi(inductiva.api.get_client())
    storage_total_size_bytes = api.get_storage_size().body
    estimated_storage_cost = api.get_storage_monthly_cost(
    ).body["estimated_monthly_cost"]

    estimated_storage_cost = format_utils.currency_formatter(
        estimated_storage_cost)
    storage_total_size = format_utils.bytes_formatter(storage_total_size_bytes)

    print("Total storage size used:")
    print(f"\tVolume: {storage_total_size}")
    print(f"\tCost: {estimated_storage_cost}/month")
    print("")

    return storage_total_size_bytes


def get_space_used():
    """Returns the occupied storage size in GB."""
    storage_total_size_bytes = _print_storage_size_and_cost()
    #Return float instead of a string
    storage_used = round(float(storage_total_size_bytes) / (1024**3), 3)
    return storage_used


def listdir(path="/",
            max_results: int = 10,
            order_by: Literal["size", "creation_time"] = "creation_time",
            sort_order: Literal["asc", "desc"] = "desc"):
    """List and display the contents of the user's storage.
    Args:
        path (str): Storage directory to list. Default is root.
        max_results (int): The maximum number of results to return.
        order_by (str): The field to sort the contents by.
        sort_order (str): Whether to sort the contents in ascending or
        descending order.
    Returns:
        list of dict: A list of dictionaries containing information about
        the size, the name and the creation time of each content that can
        easily be converted to a dataframe.

    This function prints a table with the storage content information:
        Name            Size            Creation Time
        1234            5.68 MiB        29 Sep, 14:12:00
        12345           374.85 KiB      29 Sep, 14:13:10
        1234567         97.59 KiB       29 Sep, 14:13:24
        123             0 B             29 Sep, 14:13:29
        123456          0 B             29 Sep, 14:13:18
    You can use this information to delete the contents you don't need
    anymore and further inspect task outputs and logs using the Task
    class.
    """

    api = storage_api.StorageApi(inductiva.api.get_client())

    # This is valid for single files and directories
    if len(path.split(os.sep)) < 2:
        path += os.sep

    contents = api.list_storage_contents({
        "path": path,
        "max_results": max_results,
        "sort_by": order_by,
        "order": sort_order
    }).body
    all_contents = []
    for content_name, info in contents.items():
        size = info["size_bytes"]
        creation_time = info["creation_time"]
        all_contents.append({
            "content_name": content_name,
            "size": round(float(size), 3),
            "creation_time": creation_time
        })
    print(_print_contents_table(all_contents))

    _print_storage_size_and_cost()

    print(f"Listed {len(all_contents)} folder(s). Ordered by {order_by}.\n"
          "Use --max-results/-m to control the number of results displayed.")
    return all_contents


def _print_contents_table(contents):
    columns = ["Name", "Size", "Creation Time"]
    rows = []

    for content in contents:
        row = [
            content["content_name"], content["size"], content["creation_time"]
        ]
        rows.append(row)

    formatters = {
        "Creation Time": [format_utils.datetime_formatter],
        "Size": [format_utils.bytes_formatter],
    }

    emph_formatter = format_utils.get_ansi_formatter()
    header_formatters = [
        lambda x: emph_formatter(x.upper(), format_utils.Emphasis.BOLD)
    ]

    return format_utils.get_tabular_str(
        rows,
        columns,
        formatters=formatters,
        header_formatters=header_formatters,
    )


def upload_from_url(
    url: str,
    remote_dir: str,
    file_name: Optional[str] = None,
):
    """
    Upload a file from a given URL to a specified remote directory.

    If no file name is provided, the function extracts the name from the URL.

    Args:
        url (str): The URL of the file to be uploaded.
        remote_dir (str): The path to the remote directory where the file will
            be stored.
        file_name (str, optional): The name to save the uploaded file as.
            If not specified, the name will be extracted from the URL.
    """
    api_instance = storage_api.StorageApi(inductiva.api.get_client())

    # append filename extracted from url to remote_path
    if not file_name:
        parsed_url = urllib.parse.urlparse(url)
        file_name = urllib.parse.unquote(parsed_url.path.split("/")[-1])

    remote_path = os.path.join(remote_dir, file_name)

    api_instance.upload_from_url(query_params={
        "url": url,
        "path": remote_path,
        "unzip": "f",
    },)
    logging.info("File is being uploaded...")
    logging.info("You can use 'inductiva storage ls' to check the status.")


def upload(
    local_path: str,
    remote_dir: str,
):
    """
    Upload a local file or directory to a specified remote directory.

    Args:
        local_path (str): The path to the local file or directory to be
            uploaded.
        remote_dir (str, optional): The remote directory where the file will
            be uploaded.
    """
    is_dir = os.path.isdir(local_path)

    if is_dir:
        local_dir = os.path.join(local_path, "")
        file_paths, total_size = _list_files(local_path)
        remote_file_paths = [
            os.path.join(remote_dir, file_path) for file_path in file_paths
        ]
    else:
        local_dir = os.path.dirname(local_path)
        filename = os.path.basename(local_path)
        remote_file_paths = [os.path.join(remote_dir, filename)]
        total_size = os.path.getsize(local_path)

    if os.path.join(remote_dir, constants.TASK_OUTPUT_ZIP) in remote_file_paths:
        raise ValueError(f"Invalid file name: '{constants.TASK_OUTPUT_ZIP}.'")

    logging.info("Uploading input...")

    api_instance = storage_api.StorageApi(inductiva.api.get_client())

    api_response = methods.get_upload_url(
        api_instance.get_upload_url,
        query_params={"paths": remote_file_paths},
    )

    with tqdm.tqdm(total=total_size,
                   unit="B",
                   unit_scale=True,
                   unit_divisor=1000) as progress_bar:

        for response in api_response:
            method = response["method"]
            url = response["url"]
            remote_file_path = response["file_path"]

            file_path = remote_file_path.removeprefix(f"{remote_dir}/")
            local_file_path = os.path.join(local_dir, file_path)

            try:
                methods.upload_file(api_instance, local_file_path, method, url,
                                    progress_bar)

                methods.notify_upload_complete(
                    api_instance.notify_upload_file,
                    query_params={
                        "path": remote_file_path,
                    },
                )
            except exceptions.ApiException as e:
                raise e

    logging.info("Input uploaded successfully.")


def _list_files(root_path: str) -> Tuple[List[str], int]:
    """
    Lists all files within a directory and returns their total size.

    Args:
        root_path: The root directory to list files from.
    """
    file_paths = []
    total_size = 0

    for dirpath, _, filenames in os.walk(root_path):
        for filename in filenames:
            path = os.path.relpath(os.path.join(dirpath, filename), root_path)
            file_paths.append(path)

            full_path = os.path.join(dirpath, filename)
            total_size += os.path.getsize(full_path)

    return file_paths, total_size


def remove_workspace(remote_dir) -> bool:
    """
    Removes path from a remote directory.

    Parameters:
    - remote_dir (str): The path to the remote directory.
    """
    api = storage_api.StorageApi(inductiva.api.get_client())

    logging.info("Removing workspace file(s)...")

    # Since we don't allow root files in workspaces it must be a directory
    # otherwise path validation in the backend will give error
    if "/" not in remote_dir:
        remote_dir = remote_dir + "/"
    api.delete_file(query_params={"path": remote_dir},)
    logging.info("Workspace file(s) removed successfully.")


class StorageOperation():
    """Represents a storage operation running remotely via Inductiva API."""

    def __init__(self, api, id_):
        self._api = api
        self.id = id_

    def _update_from_api_response(self, response):
        self._name = response["name"]
        self._status = response["status"]
        self._attributes = response["attributes"]
        self._start_time = response["start_time"]
        self._end_time = response["end_time"]
        self._error_message = response["error_message"]

    @classmethod
    def from_api_response(cls, api, response):

        op = cls(api, response["id"])

        op._update_from_api_response(response,)
        return op

    def _refresh(self):
        resp = self._api.get_operation(path_params={
            "operation_id": self.id
        }).body

        self._update_from_api_response(resp)

    def wait(self, poll_s: int = 2):
        """Wait for the operation to complete.

        Args:
            poll_s: Time in seconds between calls to the API to update
                the status of the operation.
        """

        while self._status == models.OperationStatus.RUNNING:
            self._refresh()
            time.sleep(poll_s)

        if self._status == models.OperationStatus.FAILED:
            logging.error("Operation failed: %s", self._error_message)
        else:
            logging.info("Operation completed successfully.")

        return self._status


def export(path: str, dest_url: str) -> StorageOperation:
    """Export files from the API's storage to a remote storage location.

    Args:
        path: Path in the API's remote storage to the files.
        dest_url: URL to upload the output files.

    Returns:
        Instance of StorageOperation. Call `wait` on the resulting object
        to block until the operation finishes.
    """
    api = storage_api.StorageApi(inductiva.api.get_client())
    resp = api.export_files(body={
        "path": path,
        "dest_url": dest_url,
    })

    logging.info("Started export operation ...")
    operation = StorageOperation.from_api_response(api, resp.body)

    return operation



"""NWChem simulator module of the API."""

from typing import List, Optional

from inductiva import types, tasks, simulators
from inductiva.commands.commands import Command
from inductiva.commands.mpiconfig import MPIConfig


@simulators.simulator.mpi_enabled
class NWChem(simulators.Simulator):
    """Class to invoke a generic NWChem simulation on the API."""

    def __init__(self, /, version: Optional[str] = None, use_dev: bool = False):
        """Initialize the NWChem simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "nwchem"
        self.container_image = self._get_image_uri()

    def run(self,
            input_dir: Optional[str],
            sim_config_filename: str,
            *,
            on: types.ComputationalResources,
            n_vcpus: Optional[int] = None,
            use_hwthread: bool = True,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[List[str]] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            on: The computational resource to launch the simulation on.
            sim_config_filename: Name of the simulation configuration file.
            n_vcpus: Number of vCPUs to use in the simulation. If not provided
            (default), all vCPUs will be used.
            use_hwthread: If specified Open MPI will attempt to discover the
            number of hardware threads on the node, and use that as the
            number of slots available.
            other arguments: See the documentation of the base class.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
        """
        mpi_kwargs = {}
        mpi_kwargs["use_hwthread_cpus"] = use_hwthread
        if n_vcpus is not None:
            mpi_kwargs["np"] = n_vcpus

        mpi_config = MPIConfig(version="4.1.6", **mpi_kwargs)
        commands = [
            Command(f"nwchem {sim_config_filename}", mpi_config=mpi_config)
        ]

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           remote_assets=remote_assets,
                           resubmit_on_preemption=resubmit_on_preemption,
                           **kwargs)



"""Loader module to register all modules within each package automatically."""
import importlib
import pkgutil
import logging

_logger = logging.getLogger(__name__)


def load_commands(parser, module_dir, package, prefix="", ignores_prefix=None):
    """Load all commands within the module_dir and package."""
    modules = pkgutil.iter_modules([module_dir])
    for _, name, _ in modules:
        if not name.startswith(prefix) or \
            (ignores_prefix is not None and name.startswith(ignores_prefix)):
            continue

        modname = f".{name}"
        try:
            module = importlib.import_module(modname, package)
        except Exception as e:  # pylint: disable=broad-except
            _logger.error("Failed to import %s: %s", modname, e)
            continue

        if hasattr(module, "register"):
            indent = package.count(".") * "  "
            _logger.debug("%sRegistering %s", indent, modname)
            module.register(parser)



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class CostDetail(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "components",
            "total",
            "month",
            "type",
        }

        class properties:
            month = schemas.StrSchema
            total = schemas.NumberSchema

            @staticmethod
            def components() -> typing.Type['CostComponents']:
                return CostComponents

            @staticmethod
            def type() -> typing.Type['CostType']:
                return CostType

            class warning(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'warning':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            __annotations__ = {
                "month": month,
                "total": total,
                "components": components,
                "type": type,
                "warning": warning,
            }

    components: 'CostComponents'
    total: MetaOapg.properties.total
    month: MetaOapg.properties.month
    type: 'CostType'

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["month"]
    ) -> MetaOapg.properties.month:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["total"]
    ) -> MetaOapg.properties.total:
        ...

    @typing.overload
    def __getitem__(
            self,
            name: typing_extensions.Literal["components"]) -> 'CostComponents':
        ...

    @typing.overload
    def __getitem__(self,
                    name: typing_extensions.Literal["type"]) -> 'CostType':
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["warning"]
    ) -> MetaOapg.properties.warning:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "month",
        "total",
        "components",
        "type",
        "warning",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["month"]
    ) -> MetaOapg.properties.month:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["total"]
    ) -> MetaOapg.properties.total:
        ...

    @typing.overload
    def get_item_oapg(
            self,
            name: typing_extensions.Literal["components"]) -> 'CostComponents':
        ...

    @typing.overload
    def get_item_oapg(self,
                      name: typing_extensions.Literal["type"]) -> 'CostType':
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["warning"]
    ) -> typing.Union[MetaOapg.properties.warning, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "month",
        "total",
        "components",
        "type",
        "warning",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        components: 'CostComponents',
        total: typing.Union[
            MetaOapg.properties.total,
            decimal.Decimal,
            int,
            float,
        ],
        month: typing.Union[
            MetaOapg.properties.month,
            str,
        ],
        type: 'CostType',
        warning: typing.Union[MetaOapg.properties.warning, dict,
                              frozendict.frozendict, str, date, datetime,
                              uuid.UUID, int, float, decimal.Decimal, bool,
                              None, list, tuple, bytes, io.FileIO,
                              io.BufferedReader, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'CostDetail':
        return super().__new__(
            cls,
            *_args,
            components=components,
            total=total,
            month=month,
            type=type,
            warning=warning,
            _configuration=_configuration,
            **kwargs,
        )


from inductiva.client.model.cost_components import CostComponents
from inductiva.client.model.cost_type import CostType



"""Methods to interact with the user info on the API."""
import json
from typing import Any, Dict

from inductiva import api
from inductiva.client.apis.tags.users_api import UsersApi


def _fetch_quotas_from_api() -> Dict[str, Dict[str, Any]]:
    """Get information about a user's quotas.
    """

    with api.get_client() as client:
        api_instance = UsersApi(client)

        resp = api_instance.get_user_quotas(skip_deserialization=True).response
        quotas = json.loads(resp.data.decode("utf-8"))

        return quotas


def get_quotas() -> Dict[str, Dict[str, Any]]:
    """Get the user quotas.

    This function gets a dict with the user quotas.

    Returns:
        Dict with the user quotas.
    """
    quotas = _fetch_quotas_from_api()

    return quotas


def get_info() -> Dict[str, Any]:
    """Get the user information.

    This funtion gets the user information, including the user's name, email,
    username, tier, programs, and total available credits.

    Returns:
        Dict with the user information.
    """

    with api.get_client() as client:
        api_instance = UsersApi(client)
        request = api_instance.get_user_info()
    return request.body


def get_costs(start_year: int,
              start_month: int,
              end_year: int = None,
              end_month: int = None) -> Dict[str, Any]:
    """Get the user costs.

    This function gets a dict with the user costs.

    Returns:
        Dict with the user costs.
    """
    if end_year and not end_month:
        raise ValueError("If end_year is provided, "
                         "end_month must also be provided.")
    if end_month and not end_year:
        raise ValueError("If end_month is provided, "
                         "end_year must also be provided.")

    with api.get_client() as client:
        api_instance = UsersApi(client)

        query_params = {"start_year": start_year, "start_month": start_month}

        #add end year and month if provided
        if end_year and end_month:
            query_params["end_year"] = end_year
            query_params["end_month"] = end_month

        request = api_instance.get_user_costs(query_params=query_params)
    return request.body["costs"]



# coding: utf-8
"""


    Generated by: https://openapi-generator.tech
"""

from dataclasses import dataclass
import typing_extensions
import urllib3
from urllib3._collections import HTTPHeaderDict

from inductiva.client import api_client, exceptions
from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401

from inductiva.client.model.capability import Capability

from . import path

_auth = [
    'APIKeyHeader',
]


class SchemaFor200ResponseBodyApplicationJson(schemas.ListSchema):

    class MetaOapg:

        @staticmethod
        def items() -> typing.Type['Capability']:
            return Capability

    def __new__(
        cls,
        _arg: typing.Union[typing.Tuple['Capability'],
                           typing.List['Capability']],
        _configuration: typing.Optional[schemas.Configuration] = None,
    ) -> 'SchemaFor200ResponseBodyApplicationJson':
        return super().__new__(
            cls,
            _arg,
            _configuration=_configuration,
        )

    def __getitem__(self, i: int) -> 'Capability':
        return super().__getitem__(i)


@dataclass
class ApiResponseFor200(api_client.ApiResponse):
    response: urllib3.HTTPResponse
    body: typing.Union[
        SchemaFor200ResponseBodyApplicationJson,
    ]
    headers: schemas.Unset = schemas.unset


_response_for_200 = api_client.OpenApiResponse(
    response_cls=ApiResponseFor200,
    content={
        'application/json':
            api_client.MediaType(schema=SchemaFor200ResponseBodyApplicationJson
                                ),
    },
)
_status_code_to_response = {
    '200': _response_for_200,
}
_all_accept_content_types = ('application/json',)


class BaseApi(api_client.Api):

    @typing.overload
    def _get_user_capabilities_oapg(
        self,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
            ApiResponseFor200,
    ]:
        ...

    @typing.overload
    def _get_user_capabilities_oapg(
        self,
        skip_deserialization: typing_extensions.Literal[True],
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
    ) -> api_client.ApiResponseWithoutDeserialization:
        ...

    @typing.overload
    def _get_user_capabilities_oapg(
        self,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = ...,
    ) -> typing.Union[
            ApiResponseFor200,
            api_client.ApiResponseWithoutDeserialization,
    ]:
        ...

    def _get_user_capabilities_oapg(
        self,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = False,
    ):
        """
        Get User Capabilities
        :param skip_deserialization: If true then api_response.response will be set but
            api_response.body and api_response.headers will not be deserialized into schema
            class instances
        """
        used_path = path.value

        _headers = HTTPHeaderDict()
        # TODO add cookie handling
        if accept_content_types:
            for accept_content_type in accept_content_types:
                _headers.add('Accept', accept_content_type)

        response = self.api_client.call_api(
            resource_path=used_path,
            method='get'.upper(),
            headers=_headers,
            auth_settings=_auth,
            stream=stream,
            timeout=timeout,
        )

        if skip_deserialization:
            api_response = api_client.ApiResponseWithoutDeserialization(
                response=response)
        else:
            response_for_status = _status_code_to_response.get(
                str(response.status))
            if response_for_status:
                api_response = response_for_status.deserialize(
                    response, self.api_client.configuration)
            else:
                api_response = api_client.ApiResponseWithoutDeserialization(
                    response=response)

        if not 200 <= response.status <= 299:
            raise exceptions.ApiException(status=response.status,
                                          reason=response.reason,
                                          api_response=api_response)

        return api_response


class GetUserCapabilities(BaseApi):
    # this class is used by api classes that refer to endpoints with operationId fn names

    @typing.overload
    def get_user_capabilities(
        self,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
            ApiResponseFor200,
    ]:
        ...

    @typing.overload
    def get_user_capabilities(
        self,
        skip_deserialization: typing_extensions.Literal[True],
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
    ) -> api_client.ApiResponseWithoutDeserialization:
        ...

    @typing.overload
    def get_user_capabilities(
        self,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = ...,
    ) -> typing.Union[
            ApiResponseFor200,
            api_client.ApiResponseWithoutDeserialization,
    ]:
        ...

    def get_user_capabilities(
        self,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = False,
    ):
        return self._get_user_capabilities_oapg(
            accept_content_types=accept_content_types,
            stream=stream,
            timeout=timeout,
            skip_deserialization=skip_deserialization)


class ApiForget(BaseApi):
    # this class is used by api classes that refer to endpoints by path and http method names

    @typing.overload
    def get(
        self,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
            ApiResponseFor200,
    ]:
        ...

    @typing.overload
    def get(
        self,
        skip_deserialization: typing_extensions.Literal[True],
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
    ) -> api_client.ApiResponseWithoutDeserialization:
        ...

    @typing.overload
    def get(
        self,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = ...,
    ) -> typing.Union[
            ApiResponseFor200,
            api_client.ApiResponseWithoutDeserialization,
    ]:
        ...

    def get(
        self,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = False,
    ):
        return self._get_user_capabilities_oapg(
            accept_content_types=accept_content_types,
            stream=stream,
            timeout=timeout,
            skip_deserialization=skip_deserialization)



"""Removes the Task-Runner via CLI."""
from typing import TextIO
import argparse
import sys
from inductiva import _cli

_docker_imported = True
try:
    import docker
    from docker.errors import DockerException
except ImportError:
    _docker_imported = False


def remove_task_runner(_, fout: TextIO = sys.stdout):
    """Removes the Task-Runner."""
    if not _docker_imported:
        print(
            "Docker Python API not installed, please run "
            "'pip install inductiva[task-runner]' to install it",
            file=fout)
        return

    try:
        client = docker.from_env()
    except DockerException as e:
        print(f"Failed to connect to Docker: {e}", file=fout)
        print(
            "Please make sure Docker is running "
            "and you have the necessary permissions.",
            file=fout)
        return

    task_runner = client.containers.list(filters={"name": "task-runner"})
    file_tracker = client.containers.list(filters={"name": "file-tracker"})

    if not (task_runner or file_tracker):
        print("Task-Runner is already stopped", file=fout)
        return

    if file_tracker:
        print("Removing the File-Tracker container.", file=fout)
        file_tracker[0].stop()
    if task_runner:
        print("Removing the Task-Runner container.", file=fout)
        task_runner[0].stop()


def register(parser):
    """Register the remove task-runner command."""
    subparser = parser.add_parser("remove",
                                  help="Removes the Task-Runner.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "The `inductiva task-runner remove` command provides "
        "a way to remove the Task-Runner running in the background.")

    _cli.utils.add_watch_argument(subparser)
    subparser.set_defaults(func=remove_task_runner)



"""Custom logging functions"""
import traceback
import json
import os

import logging.handlers
import logging
import sys

import inductiva
from inductiva import constants
from inductiva.client import exceptions
from inductiva.utils import format_utils

root_logger = logging.getLogger()


def is_cli():
    """Determines if the caller is the CLI"""
    caller = sys.argv[0]
    return caller.endswith(("inductiva", "inductiva.exe"))


def get_logs_file_path():
    return inductiva.constants.LOGS_FILE_PATH


class NoExceptionFormatter(logging.Formatter):

    def format(self, record):
        record.exc_text = record.exc_info = None
        return super().format(record)


def handle_uncaught_exception(exc_type, exc_value, exc_traceback):
    if _handle_api_exception(exc_type,
                             exc_value,
                             exc_traceback,
                             is_notebook=False):
        return
    sys.__excepthook__(exc_type, exc_value, exc_traceback)


def _get_traceback_first_and_last_lines(exc_traceback, n_lines: int,
                                        is_notebook: bool):
    """
    This method gets the first and last N lines of the traceback.
    Args:
        exc_traceback: traceback object.
        n_lines: number of lines to get.
    """
    tb_list = traceback.extract_tb(exc_traceback)

    skip = 1 if is_notebook else 0

    tb_first_list = tb_list[skip:n_lines + skip]
    tb_last_list = tb_list[-n_lines:]

    first_formatted_tb = [
        f"  - {frame.filename} line {frame.lineno}, in {frame.name}"
        for frame in tb_first_list
    ]
    last_formatted_tb = [
        f"  - {frame.filename} line {frame.lineno}, in {frame.name}"
        for frame in tb_last_list
    ]

    result_string = ("\n".join(first_formatted_tb) + "\n" + \
                     "    ...\n" +
                     "\n".join(last_formatted_tb) + "\n")

    return result_string


def _handle_api_exception(exc_type, exc_value, exc_traceback,
                          is_notebook: bool):
    if issubclass(exc_type, exceptions.ApiException) and \
        400 <= exc_value.status  < 500:
        detail = json.loads(exc_value.body)["detail"]

        # Gets the last N lines of the traceback
        formatted_tb = _get_traceback_first_and_last_lines(
            exc_traceback,
            constants.EXCEPTIONS_MAX_TRACEBACK_DEPTH,
            is_notebook=is_notebook)

        if not is_cli():
            detail = (f"{detail}\n  in:\n{formatted_tb}\n"
                      "For more information on this error, "
                      f"check the logs at {get_logs_file_path()}")

        root_logger.error("ERROR: %s",
                          detail,
                          exc_info=(exc_type, exc_value, exc_traceback))
        return True
    if issubclass(exc_type, exceptions.ApiValueError):
        root_logger.error("Error: %s",
                          exc_value,
                          exc_info=(exc_type, exc_value, exc_traceback))
        return True
    if issubclass(exc_type, inductiva.VersionError):
        root_logger.error(exc_value)
        return True

    return False


def ipy_handle_uncaught_exception(self,
                                  exc_type,
                                  exc_value,
                                  exc_traceback,
                                  tb_offset=None):
    if _handle_api_exception(exc_type,
                             exc_value,
                             exc_traceback,
                             is_notebook=True):
        return
    self.showtraceback((exc_type, exc_value, exc_traceback),
                       tb_offset=tb_offset)


def setup(level=logging.INFO):
    formatters = [
        logging.Formatter(
            fmt=
            "%(asctime)s|%(levelname)s|%(name)s|%(filename)s:%(lineno)"\
            "d %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S"),
        NoExceptionFormatter(fmt="%(message)s")
    ]

    logs_file_path = get_logs_file_path()
    logs_file_dir = logs_file_path.parent
    os.makedirs(logs_file_dir, exist_ok=True)

    handlers = [
        logging.handlers.RotatingFileHandler(logs_file_path,
                                             encoding="utf8",
                                             maxBytes=1e6,
                                             backupCount=10),
        logging.StreamHandler()
    ]

    for handler, formatter in zip(handlers, formatters):
        handler.setFormatter(formatter)
        root_logger.addHandler(handler)

    root_logger.setLevel(level)

    api_traceback = format_utils.getenv_bool("INDUCTIVA_DEBUG_API_TRACEBACK",
                                             False)
    if api_traceback:
        return

    has_ipython = False
    try:
        ipython = __import__("IPython")
        has_ipython = True
    except ImportError:
        pass

    if has_ipython and (ip := ipython.get_ipython()):
        ip.set_custom_exc((Exception,), ipy_handle_uncaught_exception)
    else:
        sys.excepthook = handle_uncaught_exception



"""Functions for running simulations via Inductiva Web API."""
import pathlib
from typing import Any, List, Optional

import logging

from inductiva import tasks, types
from inductiva.api import methods


def run_simulation(
    simulator: str,
    input_dir: Optional[pathlib.Path],
    *,
    computational_resources: types.ComputationalResources,
    resubmit_on_preemption: bool = False,
    storage_dir: Optional[str] = "",
    api_invoker=None,
    simulator_obj=None,
    input_resources: Optional[List[str]] = None,
    simulator_name_alias: Optional[str] = None,
    **kwargs: Any,
) -> tasks.Task:
    """Run a simulation via Inductiva Web API."""

    params = {
        "sim_dir": input_dir,
        **kwargs,
    }
    type_annotations = {
        "sim_dir": pathlib.Path,
    }

    if api_invoker is None:
        api_invoker = methods.invoke_async_api

    if not input_resources:
        input_resources = []

    container_image = kwargs.get("container_image", None)

    task_id = api_invoker(simulator,
                          params,
                          type_annotations,
                          computational_resources,
                          resubmit_on_preemption=resubmit_on_preemption,
                          simulator_name_alias=simulator_name_alias,
                          container_image=container_image,
                          storage_path_prefix=storage_dir,
                          simulator_obj=simulator_obj,
                          input_resources=input_resources)
    logging.info("■ Task %s submitted to the queue of the %s.", task_id,
                 computational_resources)

    if not isinstance(task_id, str):
        raise RuntimeError(
            f"Expected result to be a string with task_id, got {type(task_id)}")
    task = tasks.Task(task_id)

    position = task.get_position_in_queue()
    if position is not None:
        pos_info = f"Number of tasks ahead in the queue: {position}."
    else:
        pos_info = f"Task {task_id} does not have queue information."

    logging.info(
        "%s\n"
        "· Consider tracking the status of the task via CLI:"
        "\n\tinductiva tasks list --id %s\n"
        "· Or, tracking the logs of the task via CLI:"
        "\n\tinductiva logs %s\n"
        "· You can also get more information "
        "about the task via the CLI command:"
        "\n\tinductiva tasks info %s\n\n", pos_info, task_id, task_id, task_id)

    return task



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from inductiva.client.paths.users_organization_costs.get import GetOrganizationCosts
from inductiva.client.paths.users_capabilities.get import GetUserCapabilities
from inductiva.client.paths.users_costs.get import GetUserCosts
from inductiva.client.paths.users_info.get import GetUserInfo
from inductiva.client.paths.users_quotas.get import GetUserQuotas


class UsersApi(
        GetOrganizationCosts,
        GetUserCapabilities,
        GetUserCosts,
        GetUserInfo,
        GetUserQuotas,
):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    pass



"""Register CLI commands for logs."""
import argparse
import os

from .. import loader, utils
from ... import constants


def register(root_parser):
    parser = root_parser.add_parser(
        "logs",
        help="Stream the logs of a running task.",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.description = (
        "Stream the STDOUT of a running task.\n\n"
        "This command will stream the logs being written to STDOUT by a\n"
        "running task only if the task is still running. If the task has\n"
        "finished, or transitioned to a status other than 'running', the\n"
        "logs will not be available for streaming.\n"
        "To check the status of a task, use:\n"
        "  inductiva tasks list --id <task_id>")

    utils.show_help_msg(parser)

    loader.load_commands(parser,
                         os.path.dirname(__file__),
                         package=__name__,
                         ignores_prefix=constants.LOADER_IGNORE_PREFIX)



"""Register CLI commands for storage."""
import argparse
import os

from inductiva import constants
from inductiva._cli import loader, utils


def register(root_parser):

    parser = root_parser.add_parser(
        "task-runner",
        help="Task-Runner management utilities.",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.description = (
        "Task-Runner management utilities.\n\n"
        "The `inductiva task-runner` command allows you to "
        "manage your local task-runners on the platform.\n"
        "It provides utilities for launching or terminating task-runners.\n")

    utils.show_help_msg(parser)

    subparsers = parser.add_subparsers(title="available subcomands")
    loader.load_commands(subparsers,
                         os.path.dirname(__file__),
                         package=__name__,
                         ignores_prefix=constants.LOADER_IGNORE_PREFIX)



# do not import all endpoints into this module because that uses a lot of memory and stack frames
# if you need the ability to import all endpoints from this module, import them with
# from inductiva.client.apis.tag_to_api import tag_to_api

import enum


class TagValues(str, enum.Enum):
    COMPUTE = "compute"
    PROJECTS = "projects"
    STORAGE = "storage"
    TASKS = "tasks"
    USERS = "users"
    VERSION = "version"



# pylint: skip-file
# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

sys.path.insert(0, os.path.abspath('..'))

# -- Project information -----------------------------------------------------

project = 'Inductiva API Python client'
copyright = '2024, Inductiva Research Labs'
author = 'Inductiva Research Labs'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc', 'sphinx.ext.mathjax', 'sphinx.ext.napoleon',
    'myst_parser', 'sphinxcontrib.mermaid', 'sphinx_copybutton',
    'sphinx_tabs.tabs', 'sphinx_togglebutton', 'sphinxcontrib.googleanalytics',
    'sphinxext.opengraph', 'sphinx_sitemap', 'sphinx_reredirects'
]

myst_enable_extensions = [
    "amsmath",
    "dollarmath",
    # other MyST extensions
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']
sphinx_tabs_valid_builders = ['html']

# The suffix(es) of source filenames.
# Note: important to list ipynb before md here: we have both md and ipynb
# copies of each notebook, and myst will choose which to convert based on
# the order in the source_suffix list. Notebooks which are not executed have
# outputs stored in ipynb but not in md, so we must convert the ipynb.
source_suffix = ['.rst', '.ipynb', '.md']

# The main toctree document.
main_doc = 'index'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    'README.md',
    'markdown_sample.md',
    'task_state_diagram.md',
]

myst_links_external_new_tab = True

# Auto generate header anchors
myst_heading_anchors = 3

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_book_theme'

html_theme_options = {
    "logo": {
        "image_light": "_static/inductiva-logo-black.svg",
        "image_dark": "_static/inductiva-logo-white.svg"
    },
    "collapse_navigation": False,
    "show_nav_level": 1
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_favicon = 'favicon.ico'

html_title = 'Python API for Cloud-based Simulation'

# Google Analytics
googleanalytics_id = "G-N6GVDZ3621"
googleanalytics_enabled = True

# OpenGraph protocol
ogp_site_name = "Inductiva.ai Tutorials"
ogp_site_url = "https://tutorials.inductiva.ai"
ogp_image = "https://tutorials.inductiva.ai/_static/inductiva-social-banner.jpg"

# sitemap.xml
# See https://sphinx-sitemap.readthedocs.io/
language = 'en'
version = 'local'
html_baseurl = 'https://tutorials.inductiva.ai/'

redirects = {
    "intro_to_api/how_it_works.html":
        "https://docs.inductiva.ai/en/latest/intro_to_api/how_it_works.html",
    "intro_to_api/tasks.html":
        "https://docs.inductiva.ai/en/latest/intro_to_api/tasks.html",
    "intro_to_api/shared_dedicated_resources.html":
        "https://docs.inductiva.ai/en/latest/intro_to_api/shared_dedicated_resources.html",
    "intro_to_api/data_flow.html":
        "https://docs.inductiva.ai/en/latest/intro_to_api/data_flow.html",
    "intro_to_api/computational-infrastructure.html":
        "https://docs.inductiva.ai/en/latest/intro_to_api/computational-infrastructure.html",
    "intro_to_api/computational-infrastructure.html":
        "https://docs.inductiva.ai/en/latest/intro_to_api/computational-infrastructure.html",
    "intro_to_api/templating.html":
        "https://docs.inductiva.ai/en/latest/intro_to_api/templating.html",
    "intro_to_api/configuring-simulators.html":
        "https://docs.inductiva.ai/en/latest/intro_to_api/configuring-simulators.html",
    "intro_to_api/projects.html":
        "https://docs.inductiva.ai/en/latest/intro_to_api/projects.html"
}



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class UserCredits(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "amount",
        }

        class properties:
            amount = schemas.NumberSchema

            class tx_info(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.DictSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'tx_info':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            __annotations__ = {
                "amount": amount,
                "tx_info": tx_info,
            }

    amount: MetaOapg.properties.amount

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["amount"]
    ) -> MetaOapg.properties.amount:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["tx_info"]
    ) -> MetaOapg.properties.tx_info:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "amount",
        "tx_info",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["amount"]
    ) -> MetaOapg.properties.amount:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["tx_info"]
    ) -> typing.Union[MetaOapg.properties.tx_info, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "amount",
        "tx_info",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        amount: typing.Union[
            MetaOapg.properties.amount,
            decimal.Decimal,
            int,
            float,
        ],
        tx_info: typing.Union[MetaOapg.properties.tx_info, dict,
                              frozendict.frozendict, str, date, datetime,
                              uuid.UUID, int, float, decimal.Decimal, bool,
                              None, list, tuple, bytes, io.FileIO,
                              io.BufferedReader, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'UserCredits':
        return super().__new__(
            cls,
            *_args,
            amount=amount,
            tx_info=tx_info,
            _configuration=_configuration,
            **kwargs,
        )



"""List the user information via CLI."""
from collections import defaultdict
from datetime import datetime
from typing import TextIO
import argparse
import sys

from inductiva import users, _cli
from inductiva.client.exceptions import ApiException
from inductiva.users.methods import get_costs
from inductiva.utils import format_utils


def _quotas_to_dict(quotas):
    """Converts a list of quotas to two dictionaries.
    One for global quotas and one for instance quotas.
    """
    table_global = defaultdict(list)
    table_instance = defaultdict(list)
    for _, quota in quotas.items():
        current = quota["in_use"]
        max_allowed = quota["max_allowed"]
        unit = quota["unit"]
        current_str = (f"{current} {unit}" if current is not None else "N/A")
        max_allowed_str = (f"{max_allowed} {unit}"
                           if max_allowed is not None else "N/A")

        if quota["scope"] == "instance":
            table_instance[""].append(quota["label"])
            table_instance["max allowed"].append(max_allowed_str)
        else:
            table_global[""].append(quota["label"])
            table_global["current usage"].append(current_str)
            table_global["max allowed"].append(max_allowed_str)
    return table_global, table_instance


def _print_quotas(_, fout: TextIO = sys.stdout):
    """ Lists the user's quotas.

    Lists all the user's quotas and the quotas left for the user to use.
    """
    table, table_instance = _quotas_to_dict(users.get_quotas())

    emph_formatter = format_utils.get_ansi_formatter()
    header_formatters = [
        lambda x: emph_formatter(x.upper(), format_utils.Emphasis.BOLD)
    ]

    table = format_utils.get_tabular_str(table,
                                         header_formatters=header_formatters)

    table_instance = format_utils.get_tabular_str(
        table_instance, header_formatters=header_formatters)
    print("■ Global User quotas", end="")
    print(table, file=fout)
    print("■ Instance User quotas", end="")
    print(table_instance, file=fout, end="")
    return 0


def _print_credits_summary(user_info, fout: TextIO = sys.stdout):
    """Prints the user's credits information."""
    total_available_credits = format_utils.currency_formatter(
        user_info["total_available_credits"])
    print(f"■ Credits: {total_available_credits}", file=fout)


def _print_estimated_costs(fout: TextIO = sys.stdout):
    """Prints the user's estimated costs for the current month."""
    current_month = datetime.now().month
    current_year = datetime.now().year
    # This endpoint is not available for external users for now.
    # This try will be removed in the future.
    try:
        costs = get_costs(start_year=current_year, start_month=current_month)
        total_estimated_costs = format_utils.currency_formatter(
            costs[0]["total"])
        computation_estimated_costs = format_utils.currency_formatter(
            costs[0]["components"]["compute"])
        storage_estimated_costs = format_utils.currency_formatter(
            costs[0]["components"]["storage"])

        label_width = 12
        cost_width = 10

        print("■ Estimated Costs (current month):", file=fout)
        print(
            f"\t{'Computation:':<{label_width}} "
            f"{computation_estimated_costs:<{cost_width}}",
            file=fout)
        print(
            f"\t{'Storage:':<{label_width}} "
            f"{storage_estimated_costs:<{cost_width}}",
            file=fout)
        print(
            f"\t{'Total:':<{label_width}} "
            f"{total_estimated_costs:<{cost_width}}",
            file=fout)
    except ApiException as _:
        return


def get_info(_, fout: TextIO = sys.stdout):
    """ Lists the user's credits.

    Lists all the user's credits and the credits left for the user to use.
    """
    user_info = users.get_info()

    organization = user_info["organization"] or ""
    tier = user_info["tier"]
    username = user_info["username"]
    email = user_info["email"]
    name = user_info["name"] or ""

    print(f"\nOrganization: {organization}\n", file=fout)
    print(f"Name: {name}", file=fout)
    print(f"Email: {email}", file=fout)
    print(f"Username: {username}", file=fout)

    print("", file=fout)
    print(f"■ Tier: {tier}", file=fout)
    print("", file=fout)

    _print_credits_summary(user_info, fout=fout)

    _print_estimated_costs(fout=fout)

    print("", file=fout)
    _print_quotas(_, fout)
    print("", file=fout)

    return 0


def register(parser):
    """Register the user credits command."""

    subparser = parser.add_parser("info",
                                  help="List the user's information.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = ("The `inductiva user info` command provides "
                             "an overview of your tier and credits.\n"
                             "It shows the credits left for you to use.\n")

    _cli.utils.add_watch_argument(subparser)

    subparser.set_defaults(func=get_info)



"""Classes to manage different Google Cloud machine group types."""
import logging
from typing import Optional, Union
import datetime

from inductiva.resources import machine_types, machines_base


class MachineGroup(machines_base.BaseMachineGroup):
    """Class to launch and manage a group of machines in Google Cloud.

    A machine group is a collection of homogenous machines with given the
    configurations that are launched in Google Cloud.
    Note: The machine group will be available only after calling 'start' method.
    The billing will start only after the machines are started."""

    def __init__(
        self,
        machine_type: str,
        provider: Union[str, machine_types.ProviderType] = "GCP",
        num_machines: int = 1,
        threads_per_core: int = 2,
        spot: bool = True,
        data_disk_gb: int = 10,
        auto_resize_disk_max_gb: Optional[int] = None,
        max_idle_time: Optional[datetime.timedelta] = None,
        auto_terminate_ts: Optional[datetime.datetime] = None,
        register: bool = True,
    ) -> None:
        """Create a MachineGroup object.

        The register argument is used to indicate if the machine group should
        be registered or if it was already registered. If set as False on
        initialization, then, the machine group is not registered and it
        can not be started in the cloud. This serves has an helper argument for
        retrieving already registered machine groups that can be started, for
        example, when retrieving with the `machines_groups.get` method.
        Users should not set this argument.

        Args:
            machine_type: The type of GC machine to launch. Ex: "e2-standard-4".
              Check https://cloud.google.com/compute/docs/machine-resource for
              information about machine types.
            provider: The cloud provider of the machine group.
            num_machines: The number of virtual machines to launch.
            threads_per_core: The number of threads per core (1 or 2).
            spot: Whether to use spot machines.
            data_disk_gb: The size of the disk for user data (in GB).
            auto_resize_disk_max_gb: The maximum size in GB that the hard disk
                of the cloud VM can reach. If set, the disk will be
                automatically resized, during the execution of a task, when the
                free space falls below a certain threshold. This mechanism helps
                prevent "out of space" errors, that can occur when a task
                generates a quantity of output files that exceeds the size of
                the local storage. Increasing disk size during task execution
                increases the cost of local storage associated with the VM,
                therefore the user must set an upper limit to the disk size, to
                prevent uncontrolled costs. Once that limit is reached, the disk
                is no longer automatically resized, and if the task continues to
                output files, it will fail.
            max_idle_time: Time without executing any task, after which the
              resource will be terminated.
            auto_terminate_ts: Moment in which the resource will be
              automatically terminated.
        """
        if num_machines < 1:
            raise ValueError(
                "`num_machines` should be a number greater than 0.")

        super().__init__(
            provider=provider,
            register=register,
            machine_type=machine_type,
            data_disk_gb=data_disk_gb,
            max_idle_time=max_idle_time,
            threads_per_core=threads_per_core,
            auto_terminate_ts=auto_terminate_ts,
            auto_resize_disk_max_gb=auto_resize_disk_max_gb,
        )

        # Num_machines is the number of requested machines
        self.num_machines = num_machines
        #Number of active machines at the time of
        #the request machine_groups.get()
        self._active_machines = 0
        self.spot = spot
        self.__is_elastic = False

        if register:
            logging.info("■ Registering MachineGroup configurations:")
            self._register_machine_group(num_vms=self.num_machines,
                                         spot=self.spot,
                                         is_elastic=self.__is_elastic)

    @classmethod
    def from_api_response(cls, resp: dict):
        machine_group = super().from_api_response(resp)
        machine_group.num_machines = int(resp["max_vms"])
        machine_group.provider = resp["provider_id"]
        machine_group.__dict__["_active_machines"] = int(resp["num_vms"])
        machine_group.__dict__["machines"] = resp["machines"]
        machine_group.spot = bool(resp["spot"])
        machine_group.register = False
        return machine_group

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}(name=\"{self.name}\")"

    def __str__(self):
        return f"Machine Group {self.name} with {self.machine_type} machines"

    def start(self, wait_for_quotas: bool = False):
        """Start the machine group.

        Args:
            wait_for_quotas: If True, the method will wait for quotas to
              become available before starting the resource.
        """

        return super().start(
            wait_for_quotas=wait_for_quotas,
            is_elastic=self.__is_elastic,
            num_vms=self.num_machines,
            spot=self.spot,
        )

    def terminate(self, verbose: bool = True):
        """Terminates all machines of the machine group."""
        return super().terminate(num_vms=self.num_machines,
                                 is_elastic=self.__is_elastic,
                                 verbose=verbose,
                                 spot=self.spot)

    def _log_machine_group_info(self):
        super()._log_machine_group_info()
        logging.info("\t· Number of machines:         %s", self.num_machines)
        logging.info("\t· Spot:                       %s", self.spot)
        self.estimate_cloud_cost()


class ElasticMachineGroup(machines_base.BaseMachineGroup):
    """Manages an elastic machine group in Google Cloud.

    An ElasticMachineGroup is a set of identical machines that can
    automatically scale based on CPU load. The group starts with a
    minimum number of machines and adjusts its size as needed scaling
    to the maximum number of machines, ensuring both optimal performance
    and cost efficiency.

    Note: The machine group becomes active after calling the 'start' method,
    and billing commences once the machines are initiated.
    """

    def __init__(
        self,
        machine_type: str,
        min_machines: int = 1,
        max_machines: int = 2,
        spot: bool = True,
        threads_per_core: int = 2,
        data_disk_gb: int = 10,
        auto_resize_disk_max_gb: Optional[int] = None,
        max_idle_time: Optional[datetime.timedelta] = None,
        auto_terminate_ts: Optional[datetime.datetime] = None,
        register: bool = True,
    ) -> None:
        """Create an ElasticMachineGroup object.

        The register argument is used to indicate if the machine group should
        be registered or if it was already registered. If set as False on
        initialization, then, the machine group is not registered and it
        can not be started in the cloud. This serves has an helper argument for
        retrieving already registered machine groups that can be started, for
        example, when retrieving with the `machines_groups.get` method.
        Users should not set this argument.

        Args:
            machine_type: The type of GC machine to launch. Ex: "e2-standard-4".
              Check https://cloud.google.com/compute/docs/machine-resource for
            more information about machine types.
            min_machines: The minimum number of available machines. This is
              a quantity of machines that will be started initially and the
              minimum available machines, even in cases of low CPU load.
            max_machines: The maximum number of machines a machine group
              can scale up to.
            spot: Whether to use spot machines.
            threads_per_core: The number of threads per core (1 or 2).
            data_disk_gb: The size of the disk for user data (in GB).
            auto_resize_disk_max_gb: The maximum size in GB that the hard disk
                of the cloud VM can reach. If set, the disk will be
                automatically resized, during the execution of a task, when the
                free space falls below a certain threshold. This mechanism helps
                prevent "out of space" errors, that can occur when a task
                generates a quantity of output files that exceeds the size of
                the local storage. Increasing disk size during task execution
                increases the cost of local storage associated with the VM,
                therefore the user must set an upper limit to the disk size, to
                prevent uncontrolled costs. Once that limit is reached, the disk
                is no longer automatically resized, and if the task continues to
                output files, it will fail.
            max_idle_time: Time without executing any task, after which the
              resource will be terminated.
            auto_terminate_ts: Moment in which the resource will be
              automatically terminated.
        """
        if min_machines < 0:
            raise ValueError(
                "`min_machines` should be a number equal or greater than 0.")

        if min_machines >= max_machines:
            raise ValueError("`max_machines` should be greater "
                             "than `min_machines`.")

        super().__init__(
            register=register,
            data_disk_gb=data_disk_gb,
            machine_type=machine_type,
            max_idle_time=max_idle_time,
            threads_per_core=threads_per_core,
            auto_terminate_ts=auto_terminate_ts,
            auto_resize_disk_max_gb=auto_resize_disk_max_gb,
        )

        self.min_machines = min_machines
        self.max_machines = max_machines
        self._active_machines = min_machines
        self.__is_elastic = True
        self.spot = spot

        if self.register:
            logging.info("■ Registering ElasticMachineGroup configurations:")
            self._register_machine_group(min_vms=self.min_machines,
                                         max_vms=self.max_machines,
                                         is_elastic=self.__is_elastic,
                                         num_vms=self._active_machines,
                                         spot=self.spot)

    @classmethod
    def from_api_response(cls, resp: dict):
        machine_group = super().from_api_response(resp)
        machine_group.spot = bool(resp["spot"])
        machine_group.max_machines = int(resp["max_vms"])
        machine_group.min_machines = int(resp["min_vms"])
        machine_group.__dict__["_active_machines"] = int(resp["num_vms"])
        machine_group.__dict__["machines"] = resp["machines"]
        return machine_group

    def active_machines_to_str(self) -> str:
        """Returns a string representation of the
        number of machines currently running.
        """
        return f"{self._active_machines}/{self.max_machines} (max)"

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}(name=\"{self.name}\")"

    def __str__(self):
        return f"Elastic Machine Group {self.name} with {self.machine_type} " \
             "machines"

    def start(self, wait_for_quotas: bool = False):
        """Start the elastic machine group.

        Args:
            wait_for_quotas: If True, the method will wait for quotas to
              become available before starting the resource.
        """

        return super().start(
            wait_for_quotas=wait_for_quotas,
            is_elastic=self.__is_elastic,
            num_vms=self.min_machines,
            min_vms=self.min_machines,
            max_vms=self.max_machines,
            spot=self.spot,
        )

    def terminate(self, verbose: bool = True):
        """Terminates all machines of the machine group."""
        return super().terminate(num_vms=self.min_machines,
                                 min_vms=self.min_machines,
                                 max_vms=self.max_machines,
                                 is_elastic=self.__is_elastic,
                                 verbose=verbose,
                                 spot=self.spot)

    def _log_machine_group_info(self):
        super()._log_machine_group_info()
        logging.info("\t· Maximum number of machines: %s", self.max_machines)
        logging.info("\t· Minimum number of machines: %s", self.min_machines)
        logging.info("\t· Spot:                       %s", self.spot)
        self.estimate_cloud_cost()



"""SWAN module of the API."""
from typing import List, Optional
from pathlib import Path

from inductiva import types, tasks, simulators
from inductiva.commands.commands import Command
from inductiva.commands.mpiconfig import MPIConfig


@simulators.simulator.mpi_enabled
class SWAN(simulators.Simulator):
    """Class to invoke a generic SWAN simulation on the API."""

    def __init__(self, /, version: Optional[str] = None, use_dev: bool = False):
        """Initialize the SWAN simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "swan"
        self.container_image = self._get_image_uri()

    def run(
        self,
        input_dir: Optional[str],
        sim_config_filename: Optional[str] = None,
        *,
        remote_assets: Optional[List[str]] = None,
        resubmit_on_preemption: bool = False,
        on: types.ComputationalResources,
        storage_dir: Optional[str] = "",
        n_vcpus: Optional[int] = None,
        use_hwthread: bool = True,
        command: str = "swanrun",
        **kwargs,
    ) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            sim_config_filename: Name of the simulation configuration file.
                Mandatory when using 'swanrun' command.
            n_vcpus: Number of vCPUs to use in the simulation. If not provided
            (default), all vCPUs will be used.
            use_hwthread: If specified Open MPI will attempt to discover the
            number of hardware threads on the node, and use that as the
            number of slots available.
            on: The computational resource to launch the simulation on.
            storage_dir: Directory for storing simulation results.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            command: The command to run the simulation. Default is 'swanrun'.
                The user can also specify 'swan.exe'.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
        """

        if command not in ("swanrun", "swan.exe"):
            raise ValueError("Invalid command. Use 'swanrun' or 'swan.exe'.")

        if sim_config_filename is None and command == "swanrun":
            raise ValueError("Simulation configuration file "
                             "(sim_config_filename) not provided.\n"
                             "When using 'swanrun' it is mandatory to provide "
                             "sim_config_filename.")

        commands = []

        path_config_filename = Path(sim_config_filename)
        working_dir = path_config_filename.parent
        config_file_only = path_config_filename.name

        # Swanrun uses internal MPI
        # we call apptainer run ... swanrun ... -mpi np
        if command == "swanrun":

            machinefile_command = Command(
                "dd if=/dev/stdin of=machinefile",
                f"localhost slots={on.available_vcpus}")

            commands.append(machinefile_command)

            mpi_flag = f"-mpi {n_vcpus}" if n_vcpus else ""

            swanrun_command = Command(
                f"swanrun -input {config_file_only} {mpi_flag}")
            commands.append(swanrun_command)

        # we call mpirun ... apptainer ... Swan.exe
        # works with clusters
        elif command == "swan.exe":

            mpi_kwargs = {}
            if n_vcpus is not None:
                mpi_kwargs["np"] = n_vcpus
            mpi_kwargs["use_hwthread_cpus"] = use_hwthread

            mpi_config = MPIConfig(version="4.1.6", **mpi_kwargs)
            swan_exe_command = Command(f"swan.exe {sim_config_filename}",
                                       mpi_config=mpi_config)
            commands.append(swan_exe_command)

        return super().run(input_dir,
                           on=on,
                           storage_dir=storage_dir,
                           commands=commands,
                           run_subprocess_dir=str(working_dir),
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           **kwargs)



"""Constants that can be set for the Inductiva client."""
import os
import pathlib
import platform

LOGS_WEBSOCKET_URL = os.environ.get("INDUCTIVA_TASK_LOGS_URL",
                                    "wss://logs.inductiva.ai")

TASK_RUNNER_IMAGE = os.environ.get("INDUCTIVA_TASK_RUNNER_IMAGE",
                                   "inductiva/task-runner:main")

FILE_TRACKER_IMAGE = os.environ.get("INDUCTIVA_FILE_TRACKER_IMAGE",
                                    "inductiva/file-tracker:main")

DEFAULT_QUEUE_MACHINE_TYPE = "c2-standard-4"

TASK_KILL_MAX_API_REQUESTS = 5

TASK_KILL_RETRY_SLEEP_SEC = 1.0

MAX_CONFIRMATION_LINES = 5

LOADER_COMMAND_PREFIX = "cmd_"

LOADER_IGNORE_PREFIX = "_"

# when printing the stack trace, how many lines to show
EXCEPTIONS_MAX_TRACEBACK_DEPTH = 2

TASK_FAILED_LINES_TO_DUMP = 10

INDUCTIVA_LOGS_WAIT_SLEEP_TIME = 1.0

system = platform.system()
if platform.system().lower() == "windows":
    HOME_DIR = pathlib.Path.home() / "AppData" / "inductiva"
elif platform.system().lower() in ["linux", "darwin"]:
    HOME_DIR = pathlib.Path.home() / ".inductiva"
else:
    raise RuntimeError(f"Current operating system {system} not supported.")

LOGS_FILE_PATH = HOME_DIR / "inductiva.log"
API_KEY_FILE_PATH = HOME_DIR / "api_key"

TASK_OUTPUT_ZIP = "output.zip"

LOGIN_MESSAGE = "Please login with `inductiva auth login`."



# coding: utf-8

# flake8: noqa

# import all models into this package
# if you have many models here with many references from one model to another this may
# raise a RecursionError
# to avoid this, import only the models that you directly need like:
# from inductiva.client.model.pet import Pet
# or import this package, but before doing it, use:
# import sys
# sys.setrecursionlimit(n)

from inductiva.client.model.autoscale_policy import AutoscalePolicy
from inductiva.client.model.backend_version import BackendVersion
from inductiva.client.model.capability import Capability
from inductiva.client.model.cost_components import CostComponents
from inductiva.client.model.cost_detail import CostDetail
from inductiva.client.model.cost_type import CostType
from inductiva.client.model.created_user import CreatedUser
from inductiva.client.model.currency_code import CurrencyCode
from inductiva.client.model.disk_resize_request import DiskResizeRequest
from inductiva.client.model.dynamic_disk_resize_config import DynamicDiskResizeConfig
from inductiva.client.model.executer import Executer
from inductiva.client.model.export_operation import ExportOperation
from inductiva.client.model.file_download_details import FileDownloadDetails
from inductiva.client.model.file_download_url import FileDownloadUrl
from inductiva.client.model.file_info import FileInfo
from inductiva.client.model.file_upload_details import FileUploadDetails
from inductiva.client.model.file_upload_url import FileUploadUrl
from inductiva.client.model.http_validation_error import HTTPValidationError
from inductiva.client.model.machine_group_cost_per_hour import MachineGroupCostPerHour
from inductiva.client.model.machine_group_costs import MachineGroupCosts
from inductiva.client.model.machine_group_costs_admin_view import MachineGroupCostsAdminView
from inductiva.client.model.machine_group_status import MachineGroupStatus
from inductiva.client.model.machine_group_terminate_reason import MachineGroupTerminateReason
from inductiva.client.model.machine_group_terminate_request import MachineGroupTerminateRequest
from inductiva.client.model.machine_group_time_stat import MachineGroupTimeStat
from inductiva.client.model.machine_group_type import MachineGroupType
from inductiva.client.model.machine_group_usage_statistics import MachineGroupUsageStatistics
from inductiva.client.model.machine_group_with_costs import MachineGroupWithCosts
from inductiva.client.model.machine_group_with_costs_admin_view import MachineGroupWithCostsAdminView
from inductiva.client.model.machine_info import MachineInfo
from inductiva.client.model.machine_operation import MachineOperation
from inductiva.client.model.machine_operation_type import MachineOperationType
from inductiva.client.model.machine_type import MachineType
from inductiva.client.model.operation_end import OperationEnd
from inductiva.client.model.operation_start import OperationStart
from inductiva.client.model.operation_status import OperationStatus
from inductiva.client.model.operation_type import OperationType
from inductiva.client.model.order import Order
from inductiva.client.model.org_status import OrgStatus
from inductiva.client.model.organization_cost_components import OrganizationCostComponents
from inductiva.client.model.organization_costs import OrganizationCosts
from inductiva.client.model.organization_create import OrganizationCreate
from inductiva.client.model.organization_update import OrganizationUpdate
from inductiva.client.model.organization_users import OrganizationUsers
from inductiva.client.model.output_archive_info import OutputArchiveInfo
from inductiva.client.model.project import Project
from inductiva.client.model.project_create import ProjectCreate
from inductiva.client.model.providers import Providers
from inductiva.client.model.quota import Quota
from inductiva.client.model.quota_scope import QuotaScope
from inductiva.client.model.storage_cost import StorageCost
from inductiva.client.model.storage_file_info import StorageFileInfo
from inductiva.client.model.storage_operation import StorageOperation
from inductiva.client.model.storage_operation_name import StorageOperationName
from inductiva.client.model.storage_sort_by import StorageSortBy
from inductiva.client.model.task import Task
from inductiva.client.model.task_connection import TaskConnection
from inductiva.client.model.task_machine_operation import TaskMachineOperation
from inductiva.client.model.task_metric_create import TaskMetricCreate
from inductiva.client.model.task_metrics import TaskMetrics
from inductiva.client.model.task_operation import TaskOperation
from inductiva.client.model.task_operation_name import TaskOperationName
from inductiva.client.model.task_position_in_queue import TaskPositionInQueue
from inductiva.client.model.task_request import TaskRequest
from inductiva.client.model.task_runner_api_connection_info import TaskRunnerAPIConnectionInfo
from inductiva.client.model.task_runner_register_info import TaskRunnerRegisterInfo
from inductiva.client.model.task_status import TaskStatus
from inductiva.client.model.task_status_code import TaskStatusCode
from inductiva.client.model.task_status_info import TaskStatusInfo
from inductiva.client.model.task_step import TaskStep
from inductiva.client.model.task_submitted_info import TaskSubmittedInfo
from inductiva.client.model.task_with_status_history import TaskWithStatusHistory
from inductiva.client.model.terms_and_conditions import TermsAndConditions
from inductiva.client.model.tier_full_info import TierFullInfo
from inductiva.client.model.tier_quota_detail import TierQuotaDetail
from inductiva.client.model.update_capabilities_actions import UpdateCapabilitiesActions
from inductiva.client.model.update_capabilities_request import UpdateCapabilitiesRequest
from inductiva.client.model.usage_statistics import UsageStatistics
from inductiva.client.model.user import User
from inductiva.client.model.user_activity import UserActivity
from inductiva.client.model.user_api_key import UserApiKey
from inductiva.client.model.user_computation_trend import UserComputationTrend
from inductiva.client.model.user_costs import UserCosts
from inductiva.client.model.user_costs_details import UserCostsDetails
from inductiva.client.model.user_costs_over_time import UserCostsOverTime
from inductiva.client.model.user_create import UserCreate
from inductiva.client.model.user_credits import UserCredits
from inductiva.client.model.user_email_organization import UserEmailOrganization
from inductiva.client.model.user_most_used_machine_types_overview import UserMostUsedMachineTypesOverview
from inductiva.client.model.user_most_used_simulators_overview import UserMostUsedSimulatorsOverview
from inductiva.client.model.user_organization_update import UserOrganizationUpdate
from inductiva.client.model.user_task_status_overview import UserTaskStatusOverview
from inductiva.client.model.user_tasks_overview import UserTasksOverview
from inductiva.client.model.user_tier_update import UserTierUpdate
from inductiva.client.model.user_update_terms_and_conditions import UserUpdateTermsAndConditions
from inductiva.client.model.vm_group_config import VMGroupConfig
from inductiva.client.model.validation_error import ValidationError
from inductiva.client.model.version_comparison_result import VersionComparisonResult



# pylint: disable=missing-module-docstring
from . import files
from . import data
from .files import download_from_url
from .authentication import (
    get_stored_api_key,
    is_valid_token,
    remove_stored_api_key,
    set_stored_api_key,
)



"""Prints a machine group information via CLI."""
from typing import TextIO
import argparse
import sys

from inductiva import resources, _cli
from inductiva.utils import format_utils


def machine_group_info(args, fout: TextIO = sys.stdout):
    """Prints a task information."""
    machine_group_name = args.name
    machine_group = resources.machine_groups.get_by_name(machine_group_name)
    machines = machine_group.__dict__["machines"]

    print(f"Showing machines of machine group: {machine_group_name}", file=fout)

    table_dict = resources.machine_groups.get_machine_dict(machines)

    formatters = {
        "Started": [format_utils.datetime_formatter,],
        "Last seen": [format_utils.datetime_formatter,],
    }

    emph_formatter = format_utils.get_ansi_formatter()

    header_formatters = [
        lambda x: emph_formatter(x.upper(), format_utils.Emphasis.BOLD)
    ]

    print(format_utils.get_tabular_str(table_dict,
                                       formatters=formatters,
                                       header_formatters=header_formatters),
          file=fout,
          end="")

    return 0


def register(parser):
    """Register the info machine group command."""
    subparser = parser.add_parser(
        "info",
        help="Prints information related to a machine group.",
        formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = ("The `inductiva resources info` command provides "
                             "information about a machine group.")

    _cli.utils.add_watch_argument(subparser)
    subparser.add_argument(
        "name",
        type=str,
        help="name of the machine group to get information about.")
    subparser.set_defaults(func=machine_group_info)



"""Tail command for task files."""
from typing import TextIO
import argparse
import sys
import asyncio

from inductiva import _cli, tasks


def tail(args: argparse.Namespace, fout: TextIO = sys.stdout):
    task_id = args.id
    task = tasks.Task(task_id)
    lines = asyncio.run(task._tail_file(args.filename, args.lines))  # pylint: disable=protected-access
    print(lines, file=fout)
    return 0


def register(parser):
    """Register the info tasks command."""
    subparser = parser.add_parser(
        "tail",
        help="Shows the last lines of a file in a task.",
        formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = ("The `inductiva tasks tail` command allows"
                             "to tail a file in a task that is running.")

    _cli.utils.add_watch_argument(subparser)
    subparser.add_argument("id",
                           type=str,
                           help="ID of the task to list directories.")
    subparser.add_argument("filename", type=str, help="File to tail.")
    subparser.add_argument("--lines",
                           "-l",
                           type=int,
                           default=10,
                           help="Number of lines to show.")
    subparser.set_defaults(func=tail)



"""Gx module of the API for numerical simulations."""
from typing import List, Optional

from inductiva import types, tasks, simulators


class Gx(simulators.Simulator):
    """Class to invoke a generic Gx simulation on the API.
    """

    def __init__(self, /, version: Optional[str] = None, use_dev: bool = False):
        """Initialize the Gx simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "gx"
        self.container_image = self._get_image_uri()

    def run(self,
            input_dir: Optional[str],
            sim_config_filename: str,
            *,
            on: types.ComputationalResources,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[List[str]] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.
        Args:
            input_dir: Path to the directory of the simulation input files.
            on: The computational resource to launch the simulation on.
            sim_config_filename: The name of the simulation configuration file.
            other arguments: See the documentation of the base class.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
        """

        commands = [f"gx {sim_config_filename}"]

        return super().run(input_dir,
                           on=on,
                           storage_dir=storage_dir,
                           commands=commands,
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           **kwargs)



"""Available machine types and their number of cores."""
from dataclasses import dataclass
from typing import List, Optional, Tuple, Union
import json

import inductiva
from inductiva.utils import format_utils
from inductiva.client import ApiException
from inductiva.client.apis.tags import compute_api


class ProviderType(format_utils.CaseInsensitiveEnum):
    """Enum to represent the provider of the machine to be launched."""
    GCP = "GCP"
    LOCAL = "LOCAL"


@dataclass(frozen=True)
class MachineTypeInfo:
    """Dataclass to represent a machine type.
    This class has all the information related to a machine type.
    """

    threads_per_core: int
    machine_type: str
    provider_id: str
    num_cpus: int
    price: float
    ram_gb: int
    region: str
    spot: bool
    zone: str

    def __getitem__(self, item):
        return getattr(self, item)


def list_available_machines(provider: Union[str, ProviderType]):
    """List all available machines types."""

    resources_available = get_available_machine_types(provider)
    machine_types = []

    for machine in resources_available:
        machine_types.append(machine["machine_type"])

    return tuple(machine_types)


def get_available_machine_types(
    provider: Union[str, ProviderType] = ProviderType.GCP,
    machine_families: Optional[List[str]] = None,
    machine_configs: Optional[List[str]] = None,
    vcpus_range: Optional[Tuple[int, int]] = None,
    memory_range: Optional[Tuple[int, int]] = None,
    price_range: Optional[Tuple[float, float]] = None,
    spot: Optional[bool] = None,
) -> List[MachineTypeInfo]:
    """Get all available machine types from a specified provider,
    allowing for multiple filtering parameters.
    
    Args:
        provider (Union[str, ProviderType]): 
            The cloud provider for which to list available machine types. 
            Defaults to `ProviderType.GCP`.
        machine_families (Optional[List[str]]): 
            A list of machine families to filter the available machine types
            (e.g., "c2", "c2d", "n4", etc).
        machine_configs (Optional[List[str]]): 
            A list of specific machine configurations to filter the results
            (e.g., "highcpu", "highmem", "standard").
        vcpus_range (Optional[Tuple[int, int]]): 
            A tuple defining the range of virtual CPUs (vCPUs) to filter the 
            machine types.
        memory_range (Optional[Tuple[int, int]]): 
            A tuple defining the range of memory (in MB) to filter the machine 
            types.
        price_range (Optional[Tuple[float, float]]): 
            A tuple defining the price range (in the respective currency) to 
            filter the machine types.
        spot (Optional[bool]): 
            If set to True, filters for spot instances; if False, filters for 
            on-demand instances.

    Returns:
        List[MachineTypeInfo]:
            A list of available machine types that match the specified criteria.
    """

    provider = ProviderType(provider)

    api_client = compute_api.ComputeApi(inductiva.api.get_client())

    query_params = {"provider_id": provider.value}

    if machine_families:
        query_params["machine_families"] = machine_families

    if machine_configs:
        query_params["machine_configs"] = machine_configs

    if vcpus_range:
        query_params["vcpus_range"] = vcpus_range

    if memory_range:
        query_params["memory_range"] = memory_range

    if price_range:
        query_params["price_range"] = price_range

    if spot is not None:
        query_params["spot"] = str(spot).lower()

    try:
        resp = api_client.list_available_machine_types(query_params).response

        response_body = json.loads(resp.data.decode("utf-8"))
        return [
            MachineTypeInfo(**machine_type) for machine_type in response_body
        ]
    except ApiException as e:
        raise e



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class OrganizationCosts(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "costs",
            "name",
            "id",
            "estimated_costs",
        }

        class properties:
            id = schemas.StrSchema
            name = schemas.StrSchema

            @staticmethod
            def costs() -> typing.Type['OrganizationCostComponents']:
                return OrganizationCostComponents

            class estimated_costs(schemas.ListSchema):

                class MetaOapg:

                    @staticmethod
                    def items() -> typing.Type['UserCostsDetails']:
                        return UserCostsDetails

                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple['UserCostsDetails'],
                                       typing.List['UserCostsDetails']],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                ) -> 'estimated_costs':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )

                def __getitem__(self, i: int) -> 'UserCostsDetails':
                    return super().__getitem__(i)

            class currency(
                    schemas.ComposedSchema,):

                class MetaOapg:

                    @classmethod
                    @functools.lru_cache()
                    def all_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            CurrencyCode,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'currency':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class consolidation_last_update_ts(
                    schemas.DateTimeBase,
                    schemas.ComposedSchema,
            ):

                class MetaOapg:
                    format = 'date-time'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'consolidation_last_update_ts':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            __annotations__ = {
                "id": id,
                "name": name,
                "costs": costs,
                "estimated_costs": estimated_costs,
                "currency": currency,
                "consolidation_last_update_ts": consolidation_last_update_ts,
            }

    costs: 'OrganizationCostComponents'
    name: MetaOapg.properties.name
    id: MetaOapg.properties.id
    estimated_costs: MetaOapg.properties.estimated_costs

    @typing.overload
    def __getitem__(
            self,
            name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["name"]
    ) -> MetaOapg.properties.name:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["costs"]
    ) -> 'OrganizationCostComponents':
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["estimated_costs"]
    ) -> MetaOapg.properties.estimated_costs:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["currency"]
    ) -> MetaOapg.properties.currency:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["consolidation_last_update_ts"]
    ) -> MetaOapg.properties.consolidation_last_update_ts:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "id",
        "name",
        "costs",
        "estimated_costs",
        "currency",
        "consolidation_last_update_ts",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self,
            name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["name"]
    ) -> MetaOapg.properties.name:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["costs"]
    ) -> 'OrganizationCostComponents':
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["estimated_costs"]
    ) -> MetaOapg.properties.estimated_costs:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["currency"]
    ) -> typing.Union[MetaOapg.properties.currency, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["consolidation_last_update_ts"]
    ) -> typing.Union[MetaOapg.properties.consolidation_last_update_ts,
                      schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "id",
        "name",
        "costs",
        "estimated_costs",
        "currency",
        "consolidation_last_update_ts",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        costs: 'OrganizationCostComponents',
        name: typing.Union[
            MetaOapg.properties.name,
            str,
        ],
        id: typing.Union[
            MetaOapg.properties.id,
            str,
        ],
        estimated_costs: typing.Union[
            MetaOapg.properties.estimated_costs,
            list,
            tuple,
        ],
        currency: typing.Union[MetaOapg.properties.currency, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, bool,
                               None, list, tuple, bytes, io.FileIO,
                               io.BufferedReader,
                               schemas.Unset] = schemas.unset,
        consolidation_last_update_ts: typing.Union[
            MetaOapg.properties.consolidation_last_update_ts, dict,
            frozendict.frozendict, str, date, datetime, uuid.UUID, int, float,
            decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO,
            io.BufferedReader, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'OrganizationCosts':
        return super().__new__(
            cls,
            *_args,
            costs=costs,
            name=name,
            id=id,
            estimated_costs=estimated_costs,
            currency=currency,
            consolidation_last_update_ts=consolidation_last_update_ts,
            _configuration=_configuration,
            **kwargs,
        )


from inductiva.client.model.currency_code import CurrencyCode
from inductiva.client.model.organization_cost_components import OrganizationCostComponents
from inductiva.client.model.user_costs_details import UserCostsDetails



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from inductiva.client.paths.storage_.delete import DeleteFile
from inductiva.client.paths.storage_export.post import ExportFiles
from inductiva.client.paths.storage_operations_operation_id.get import GetOperation
from inductiva.client.paths.storage_operations.get import GetOperationsByUser
from inductiva.client.paths.storage_signed_urls.get import GetSignedUrls
from inductiva.client.paths.storage_cost.get import GetStorageMonthlyCost
from inductiva.client.paths.storage_size.get import GetStorageSize
from inductiva.client.paths.storage_input_url.get import GetUploadUrl
from inductiva.client.paths.storage_contents.get import ListStorageContents
from inductiva.client.paths.storage_input_notify.post import NotifyUploadFile
from inductiva.client.paths.storage_input_remote.post import UploadFromUrl


class StorageApi(
        DeleteFile,
        ExportFiles,
        GetOperation,
        GetOperationsByUser,
        GetSignedUrls,
        GetStorageMonthlyCost,
        GetStorageSize,
        GetUploadUrl,
        ListStorageContents,
        NotifyUploadFile,
        UploadFromUrl,
):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    pass
